@@ -52,7 +52,7 @@ Options:
 For documentation see <https://obraz.pirx.ru/>.
 """
 
-import errno
+import contextlib
 import os
 import re
 import shutil
@@ -69,11 +69,10 @@ from typing import (
     Any,
     Callable,
     Iterable,
-    Dict,
     Sequence,
     TypeVar,
     Optional,
-    List,
+    TypedDict,
     Union,
     cast,
 )
@@ -95,7 +94,73 @@ __all__ = [
 
 PAGE_ENCODING = "UTF-8"
 
-DEFAULT_CONFIG: Dict[str, Any] = {
+
+class ConfigBase(TypedDict):
+    source: str
+    destination: str
+    include: list[str]
+    exclude: list[str]
+    exclude_patterns: list[str]
+    full_build_patterns: list[str]
+    host: str
+    port: str
+    baseurl: str
+    permalink: str
+
+
+class Config(ConfigBase, total=False):
+    time: datetime
+    drafts: bool
+    force: bool
+    trace: bool
+
+
+class File(TypedDict):
+    url: str
+    path: str
+
+
+class TemplateBase(TypedDict):
+    content: str
+
+
+class Template(TemplateBase, total=False):
+    layout: str
+
+
+class PageBase(Template):
+    url: str
+
+
+class Page(PageBase, total=False):
+    path: str
+    published: bool
+    raw_content: bool
+
+
+class PostBase(Page):
+    date: datetime
+    id: str
+
+
+class Post(PostBase, total=False):
+    next: "Post"  # type: ignore
+    previous: "Post"  # type: ignore
+    tags: list[str]
+
+
+class SiteContents(TypedDict, total=False):
+    files: list[File]
+    pages: list[Page]
+    posts: list[Post]
+    tags: dict[str, list[Post]]
+
+
+class Site(SiteContents, Config):
+    pass
+
+
+DEFAULT_CONFIG: ConfigBase = {
     "source": "./",
     "destination": "./_site",
     "include": [".htaccess"],
@@ -112,21 +177,22 @@ DEFAULT_CONFIG: Dict[str, Any] = {
     "host": "localhost",
     "port": "8000",
     "baseurl": "",
+    "permalink": "/{year}/{month}/{day}/{title}.html",
 }
 
 _quiet = False
-_loaders: list[Callable[[str, dict], Optional[dict]]] = []
-_processors: list[Callable[[dict], None]] = []
-_render_string = lambda s, _context, _site: s
-_file_filters: dict[str, Callable[[str, dict], str]] = {}
-_template_filters: dict[str, Callable[[str, dict], str]] = {}
+_loaders: list[Callable[[str, Config], Optional[SiteContents]]] = []
+_processors: list[Callable[[Site], None]] = []
+_render_string = lambda s, _context, _config: s
+_file_filters: dict[str, Callable[[str, Config], str]] = {}
+_template_filters: dict[str, Callable[[str, Config], str]] = {}
 _T = TypeVar("_T")
 
 
 def file_filter(extensions: Iterable[str]) -> Any:
     """Register a page content filter for file extensions."""
 
-    def wrapper(f: Callable[[str, dict], str]) -> Callable[[str, dict], str]:
+    def wrapper(f: Callable[[str, Config], str]) -> Callable[[str, Config], str]:
         for ext in extensions:
             _file_filters[ext] = f
         return f
@@ -137,39 +203,39 @@ def file_filter(extensions: Iterable[str]) -> Any:
 def template_filter(name: str) -> Any:
     """Register a template filter."""
 
-    def wrapper(f: Callable[[str, dict], str]) -> Callable[[str, dict], str]:
+    def wrapper(f: Callable[[str, Config], str]) -> Callable[[str, Config], str]:
         _template_filters[name] = f
         return f
 
     return wrapper
 
 
-def template_renderer(f: Callable[[str, dict, dict], str]) -> Any:
+def template_renderer(f: Callable[[str, dict[str, Any], Config], str]) -> Any:
     """Set a custom template renderer."""
     global _render_string
     _render_string = f
     return f
 
 
-def loader(f: Callable[[str, dict], Optional[dict]]) -> Any:
+def loader(f: Callable[[str, Config], Optional[SiteContents]]) -> Any:
     """Register a site source content loader."""
     _loaders.insert(0, f)
     return f
 
 
-def processor(f: Callable[[dict], None]) -> Any:
+def processor(f: Callable[[Site], None]) -> Any:
     """Register a site content processor."""
     _processors.insert(0, f)
     return f
 
 
-def generator(f: Callable[[dict], None]) -> Any:
+def generator(f: Callable[[Site], None]) -> Any:
     """Register a destination files generator for the site."""
     _processors.append(f)
     return f
 
 
-def fallback_loader(f: Callable[[str, dict], Optional[dict]]) -> Any:
+def fallback_loader(f: Callable[[str, Config], Optional[SiteContents]]) -> Any:
     _loaders.append(f)
     return f
 
@@ -212,7 +278,7 @@ def all_source_files(source: str, destination: str) -> Iterable[str]:
 
 
 def changed_files(
-    source: str, destination: str, config: Dict[str, Any], poll_interval: int = 1
+    source: str, destination: str, config: Config, poll_interval: int = 1
 ) -> Iterable[list[str]]:
     times: dict[str, float] = {}
     while True:
@@ -231,7 +297,7 @@ def changed_files(
         sleep(poll_interval)
 
 
-def is_file_visible(path: str, config: Dict[str, Any]) -> bool:
+def is_file_visible(path: str, config: Config) -> bool:
     """Check file name visibility according to site settings."""
     parts = path.split(os.path.sep)
     exclude = config.get("exclude", [])
@@ -266,22 +332,16 @@ def url2path(url: str) -> str:
 
 
 def make_dirs(path: str) -> None:
-    try:
+    with contextlib.suppress(FileExistsError):
         os.makedirs(path)
-    except OSError as e:
-        if e.errno == errno.EEXIST:
-            pass
 
 
 def remove(path: str) -> None:
-    try:
+    with contextlib.suppress(FileExistsError):
         if os.path.isdir(path):
             shutil.rmtree(path)
         else:
             os.remove(path)
-    except OSError as e:
-        if e.errno == errno.EEXIST:
-            pass
 
 
 def info(message: str) -> None:
@@ -330,23 +390,22 @@ def object_name(f: Any) -> str:
 
 @template_filter("markdownify")
 @file_filter([".md", ".markdown"])
-def markdown_filter(s: str, config: Any) -> str:
+def markdown_filter(s: str, config: Config) -> str:
     return markdown(s)
 
 
 @fallback_loader
-def load_file(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+def load_file(path: str, config: Config) -> Optional[SiteContents]:
     if not is_file_visible(path, config) or is_underscored(path):
         return None
+    file: File = {"url": path2url(path), "path": path}
     return {
-        "files": [{"url": path2url(path), "path": path}],
+        "files": [file],
     }
 
 
 @template_renderer
-def jinja2_render_string(
-    string: str, context: Dict[str, Any], config: Dict[str, Any]
-) -> str:
+def jinja2_render_string(string: str, context: dict[str, Any], config: Config) -> str:
     includes = os.path.join(config["source"], "_includes")
     env = Environment(loader=FileSystemLoader(includes))
     for name, f in _template_filters.items():
@@ -355,7 +414,7 @@ def jinja2_render_string(
     return t.render(**context)
 
 
-def read_template(path: str) -> Optional[Dict[str, Any]]:
+def read_template(path: str) -> Optional[Template]:
     with open(path, "rb") as fd:
         if fd.read(3) != b"---":
             return None
@@ -378,7 +437,7 @@ def read_template(path: str) -> Optional[Dict[str, Any]]:
 
 
 @loader
-def load_page(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+def load_page(path: str, config: Config) -> Optional[SiteContents]:
     if not is_file_visible(path, config) or is_underscored(path):
         return None
     name, suffix = os.path.splitext(path)
@@ -386,42 +445,42 @@ def load_page(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
         dst = f"{name}.html"
     else:
         dst = path
-    page = read_template(os.path.join(config["source"], path))
+    page = cast(Page, read_template(os.path.join(config["source"], path)))
     if not page:
         return None
-    page.update({"url": path2url(dst), "path": path})
+    page["url"] = path2url(dst)
+    page["path"] = path
     return {"pages": [page]}
 
 
 def read_post(
-    path: str, date: datetime, title: str, config: Dict[str, Any]
-) -> Optional[Dict[str, Any]]:
-    page = read_template(os.path.join(config["source"], path))
-    if not page:
+    path: str, date: datetime, title: str, config: Config
+) -> Optional[SiteContents]:
+    post = cast(Post, read_template(os.path.join(config["source"], path)))
+    if not post:
         return None
-    if "date" in page:
-        date = page["date"]
-    permalink = config.get("permalink", "/{year}/{month}/{day}/{title}.html")
+    if "date" in post:
+        date = post["date"]
     url_vars = {
         "year": f"{date.year:04}",
         "month": f"{date.month:02}",
         "day": f"{date.day:02}",
         "title": title,
     }
-    url = pathname2url(permalink.format(**url_vars))
-    page.update({"url": url, "path": path})
-    if "date" not in page:
+    post["url"] = pathname2url(config["permalink"].format(**url_vars))
+    post["path"] = path
+    if "date" not in post:
         date_str = "{year}-{month}-{day}".format(**url_vars)
-        page["date"] = datetime.strptime(date_str, "%Y-%m-%d")
-    page["id"] = "/{year}/{month}/{day}/{title}".format(**url_vars)
+        post["date"] = datetime.strptime(date_str, "%Y-%m-%d")
+    post["id"] = "/{year}/{month}/{day}/{title}".format(**url_vars)
     return {
-        "posts": [page],
-        "tags": dict((tag, [page]) for tag in page.get("tags", [])),
+        "posts": [post],
+        "tags": dict((tag, [post]) for tag in post.get("tags", [])),
     }
 
 
 @loader
-def load_post(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+def load_post(path: str, config: Config) -> Optional[SiteContents]:
     post_re = re.compile(
         r"(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})-" r"(?P<title>.+)"
     )
@@ -439,7 +498,7 @@ def load_post(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
 
 
 @loader
-def load_draft(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+def load_draft(path: str, config: Config) -> Optional[SiteContents]:
     if not config.get("drafts"):
         return None
     if "_drafts" not in path.split(os.path.sep):
@@ -447,21 +506,23 @@ def load_draft(path: str, config: Dict[str, Any]) -> Optional[Dict[str, Any]]:
     if not is_file_visible(path, config):
         return None
     title, _ = os.path.splitext(os.path.basename(path))
-    return read_post(path, config["time"], title, config)
+    return read_post(path, config.get("time", datetime.utcnow()), title, config)
 
 
-def render_layout(content: str, page: Dict[str, Any], site: Dict[str, Any]) -> str:
-    name = page.get("layout", "nil")
+def render_layout(content: str, template: Template, site: Site) -> str:
+    name = template.get("layout", "nil")
     if name == "nil":
         return content
     layout_file = os.path.join(site["source"], "_layouts", f"{name}.html")
     layout = read_template(layout_file)
     if not layout:
         raise Exception(f"Cannot load template: '{layout_file}'")
-    page_copy = page.copy()
+    layout_copy = cast(dict, layout.copy())
+    page_copy = cast(dict, template.copy())
     page_copy.pop("layout", None)
     page_copy.pop("content", None)
-    layout.update(page_copy)
+    layout_copy.update(page_copy)
+    layout = cast(Template, layout_copy)
     context = {
         "site": site,
         "page": layout,
@@ -471,7 +532,7 @@ def render_layout(content: str, page: Dict[str, Any], site: Dict[str, Any]) -> s
     return render_layout(content, layout, site)
 
 
-def render_page(page: Dict[str, Any], site: Dict[str, Any]) -> str:
+def render_page(page: Page, site: Site) -> str:
     context = {
         "site": site,
         "page": page,
@@ -487,9 +548,9 @@ def render_page(page: Dict[str, Any], site: Dict[str, Any]) -> str:
 
 
 @processor
-def process_posts(site: Dict[str, Any]) -> None:
+def process_posts(site: Site) -> None:
     """Sort and interlink posts."""
-    posts = site.setdefault("posts", [])
+    posts: list[Post] = site.setdefault("posts", [])
     posts.sort(key=lambda p: p["date"], reverse=True)
     n = len(posts)
     for i, post in enumerate(posts):
@@ -499,7 +560,7 @@ def process_posts(site: Dict[str, Any]) -> None:
             post["previous"] = posts[i - 1]
 
 
-def generate_page(page: Dict[str, Any], site: Dict[str, Any]) -> None:
+def generate_page(page: Page, site: Site) -> None:
     if not page.get("published", True):
         return
     url = page["url"]
@@ -515,16 +576,16 @@ def generate_page(page: Dict[str, Any], site: Dict[str, Any]) -> None:
 
 
 @generator
-def generate_pages(site: Dict[str, Any]) -> None:
+def generate_pages(site: Site) -> None:
     """Generate pages with YAML front matter."""
-    posts = site.get("posts", [])
+    posts = cast(list[Page], site.get("posts", []))
     pages = site.get("pages", [])
     for page in progress("Generating pages", posts + pages):
         generate_page(page, site)
 
 
 @generator
-def generate_files(site: Dict[str, Any]) -> None:
+def generate_files(site: Site) -> None:
     """Copy static files."""
     for file_dict in site.get("files", []):
         src = os.path.join(site["source"], file_dict["path"])
@@ -545,39 +606,39 @@ def load_plugins(source: str) -> None:
         info(f"Loaded {n} plugins")
 
 
-def build(config: Dict[str, Any]) -> None:
+def build(config: Config) -> None:
     site = load_site(config)
     generate_site(site)
 
 
-def build_delta(paths: Iterable[str], config: Dict[str, Any]) -> None:
+def build_delta(paths: Iterable[str], config: Config) -> None:
     site = load_site_files(paths, config)
     generate_site(site, clean=False)
 
 
-def load_site_files(paths: Iterable[str], config: Dict[str, Any]) -> Dict[str, Any]:
+def load_site_files(paths: Iterable[str], config: Config) -> Site:
     source = config["source"]
     info("Loading source files...")
-    site = config.copy()
+    site = cast(SiteContents, config.copy())
     n = 0
     for path in paths:
         rel_path = os.path.relpath(path, source)
         for f in _loaders:
-            data = f(rel_path, site)
+            data = f(rel_path, config)
             if data:
                 n += 1
                 site = merge(site, data)
                 break
     info(f"Loaded {n} files")
-    return site
+    return cast(Site, site)
 
 
-def load_site(config: Dict[str, Any]) -> Dict[str, Any]:
+def load_site(config: Config) -> Site:
     paths = all_source_files(config["source"], config["destination"])
     return load_site_files(paths, config)
 
 
-def generate_site(site: Dict[str, Any], clean: bool = True) -> None:
+def generate_site(site: Site, clean: bool = True) -> None:
     destination = site["destination"]
     marker = os.path.join(destination, ".obraz_destination")
     write_denied = os.path.exists(destination) and not os.path.exists(marker)
@@ -600,7 +661,7 @@ def generate_site(site: Dict[str, Any], clean: bool = True) -> None:
     info("Site generated successfully")
 
 
-def make_server(config: Dict[str, Any]) -> HTTPServer:
+def make_server(config: Config) -> HTTPServer:
     host = config["host"]
     port = int(config["port"])
     baseurl = config["baseurl"]
@@ -618,7 +679,7 @@ def make_server(config: Dict[str, Any]) -> HTTPServer:
     return HTTPServer((host, port), Handler)
 
 
-def serve(config: Dict[str, Any]) -> None:
+def serve(config: Config) -> None:
     build(config)
     server = make_server(config)
     os.chdir(config["destination"])
@@ -626,7 +687,7 @@ def serve(config: Dict[str, Any]) -> None:
     server.serve_forever()
 
 
-def watch(config: Dict[str, Any]) -> None:
+def watch(config: Config) -> None:
     source = os.path.abspath(config["source"])
     destination = os.path.abspath(config["destination"])
     initial_dir = os.getcwd()
@@ -656,14 +717,14 @@ def watch(config: Dict[str, Any]) -> None:
             serving = True
 
 
-def log_serving(config: Dict[str, Any]) -> None:
+def log_serving(config: Config) -> None:
     url = "http://{host}:{port}{baseurl}".format(**config)
     if not url.endswith("/"):
         url += "/"
     info(f"Serving at {url}")
 
 
-def full_build_required(changed_paths: Iterable[str], config: Dict[str, Any]) -> bool:
+def full_build_required(changed_paths: Iterable[str], config: Config) -> bool:
     patterns = config.get("full_build_patterns", [])
     source = os.path.abspath(config["source"])
     for path in changed_paths:
@@ -685,7 +746,7 @@ def new_site(path: str) -> None:
     info(f"New Obraz site installed in '{path}'")
 
 
-def obraz(argv: List[str]) -> None:
+def obraz(argv: list[str]) -> None:
     opts = docopt(__doc__ or "", argv=argv, version="0.9")
     global _quiet
     _quiet = opts["--quiet"]
@@ -695,14 +756,15 @@ def obraz(argv: List[str]) -> None:
             new_site(opts["PATH"])
             return
 
-        config = DEFAULT_CONFIG.copy()
+        copy = cast(dict, DEFAULT_CONFIG.copy())
         source = opts["--source"] if opts["--source"] else "./"
         config_file = os.path.join(source, "_config.yml")
-        config.update(load_yaml_mapping(config_file))
-        config["time"] = datetime.utcnow()
+        copy.update(load_yaml_mapping(config_file))
+        copy["time"] = datetime.utcnow()
         for k, v in opts.items():
             if k.startswith("--") and v:
-                config[k[2:]] = v
+                copy[k[2:]] = v
+        config = cast(Config, copy)
 
         info(f'Source: {os.path.abspath(config["source"])}')
         info(f'Destination: {os.path.abspath(config["destination"])}')
